"""
Copyright 2008-2014 by Aerospike, Inc. All rights reserved.
THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE. THE COPYRIGHT NOTICE
ABOVE DOES NOT EVIDENCE ANY ACTUAL OR INTENDED PUBLICATION.
"""


# -*- coding: utf-8 -*-

from libs import management
from helpers.encryption import TwoWayEncryption
from helpers.synchronizing import synchronous
from helpers.caching import CachedMethod
from constants import BACKUP_AND_RESTORE_RESULT_HISTORY_TIME
import uuid, datetime, sqlite3, threading, json


class BackupAndRestoreManager(object):

    def __init__(self, host, port):

        self._host = host
        self._port = port
        self._conn = sqlite3.connect(':memory:')
        self._conn.text_factory = str
        self._encryptions = dict()
        self.initialize_db()
        self.backup_progress_lock = threading.RLock()
        self.restore_init_lock = threading.RLock()
        self.restore_progress_lock = threading.RLock()
        self._restore = {}

    def initialize_db(self):

        c = self._conn.cursor()
        c.execute("""CREATE TABLE backups(
            id TEXT PRIMARY KEY,
            namespace TEXT,
            destination_node_address TEXT,
            username TEXT,
            password TEXT,
            destination_location TEXT,
            sets TEXT,
            only_metadata TEXT,
            progress_file TEXT,
            delete_it INT,
            timestamp TIMESTAMP,
            status TEXT)""")
        self._conn.commit()

    def add_backup(self, backup_id, namespace, destination_node_address, username,
         password, destination_location, sets, only_metadata, progress_file):

        if username != "" and password is not None:
            self._encryptions[backup_id] = TwoWayEncryption()
            username = self._encryptions[backup_id].encrypt(str(username))
            password = self._encryptions[backup_id].encrypt(str(password))

        c = self._conn.cursor()
        c.execute("""INSERT INTO backups VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)""",
            (backup_id, namespace, destination_node_address, username,
            password, destination_location, sets, only_metadata, progress_file, 0, None, None))

    def get_backups(self):

        c = self._conn.cursor()
        c.execute("SELECT * FROM backups")
        return c.fetchall()

    def get_successful_backups(self):

        c = self._conn.cursor()
        c.execute("SELECT destination_node_address, destination_location, namespace, sets, only_metadata from backups where status like ?", ("%Success%", ))
        return c.fetchall()

    def check_if_backup_exists(self, namespace, destination_node_address,
         destination_location, sets, only_metadata):

        c = self._conn.cursor()
        c.execute("SELECT id, count(id) FROM backups WHERE namespace like ? and destination_node_address like ? and destination_location like ? and sets like ? and only_metadata like ? and status not like ? and status not like ? and status not like ?", (namespace, destination_node_address, destination_location, sets, only_metadata, None, "%Success%", "%Failure%"))
        return c.fetchone()

    def delete_backup(self, backup_id):

        c = self._conn.cursor()
        c.execute("DELETE FROM backups WHERE id like ?", (backup_id, ))

    def update_backup_status(self, backup_id, status):

        c = self._conn.cursor()
        c.execute("UPDATE backups SET timestamp = ?, status = ? WHERE id like ?", (datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S"), status, backup_id))

    def mark_backup_to_delete_it(self, backup_id):

        c = self._conn.cursor()
        c.execute("UPDATE backups SET delete_it = 1, timestamp = ? WHERE id like ?", (datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S"), backup_id))

    def initiate_backup(self, namespace, destination_node_address, username,
    password, destination_location, aerospike_build, sets="ALL",
    only_metadata="False", scan_priority="2", terminate_on_change="True", usr=None, passwd=None, pl_password=None):

        backup_check = self.check_if_backup_exists(namespace, destination_node_address,
        destination_location, sets, only_metadata)
        if backup_check[1] != 0:
            return dict(status="Failure", error="Backup already initiated", id=backup_check[0])
        optional_args = ""
        if sets != "ALL" and sets != "":
            optional_args += "-s %s " % sets
        if only_metadata == "True":
            optional_args += "-x "
        if terminate_on_change == "True":
            optional_args += "-c "
        if scan_priority != "2":
            optional_args += "-f %s " % str(scan_priority)
        if usr and passwd:
            optional_args += "-U %s -P%s" % (usr, pl_password)

        backup_id = str(uuid.uuid4())
        progress_file = "/tmp/backup-" + backup_id + ".log"

        timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
        destination_location = destination_location.rstrip("/") + "/backup_" + timestamp
        if aerospike_build.startswith('3'):
            cmd = "/bin/bash -c 'mkdir -p %s && asbackup -h %s -p %s -d %s -n %s %s &> %s;'" % (destination_location, self._host, self._port, destination_location, namespace, optional_args, progress_file)
        else:
            cmd = "/bin/bash -c 'mkdir -p %s && clbackup2 -h %s -p %s -d %s -n %s %s &> %s;'" % (destination_location, self._host, self._port, destination_location, namespace, optional_args, progress_file)

        result = management.handle_remote_command(destination_node_address,
        username, password, cmd, background=3600, backup_flag=True,
        module_name="shell")

        if result['status'] != "Failure":
            self.add_backup(backup_id, namespace, destination_node_address,
            username, password, destination_location, sets, only_metadata,
            progress_file)
            result.update(backup_id=backup_id)
        return result

    @CachedMethod()
    @synchronous('backup_progress_lock')
    def get_backup_progress(self, cluster_level_alerts):

        backups = self.get_backups()
        response = dict()

        for backup in backups:
            if backup[9] != 1:
                response[backup[0]] = dict()
                response[backup[0]]['namespace'] = backup[1]
                response[backup[0]]['destination_node_address'] = backup[2]
                response[backup[0]]['destination_location'] = backup[5]
                if backup[11] is not None:
                    response[backup[0]]['progress'] = json.loads(backup[11])
                    timestamp = datetime.datetime.now() - datetime.timedelta(seconds=BACKUP_AND_RESTORE_RESULT_HISTORY_TIME)
                    timestamp = timestamp.strftime("%Y/%m/%d %H:%M:%S")
                    if timestamp >= backup[10]:
                        self.mark_backup_to_delete_it(backup[0])
                else:
                    username = backup[3]
                    password = backup[4]
                    if username != "" and password is not None:
                        username = self._encryptions[backup[0]].decrypt(str(username))
                        password = self._encryptions[backup[0]].decrypt(str(password))
                    response[backup[0]]['progress'] = management.get_backup_progress(
                        backup[2], username, password, backup[8], backup[1])
                    if response[backup[0]]['progress']['status'] == "Success" or response[backup[0]]['progress']['status'] == "Failure":
                        self.update_backup_status(backup[0], json.dumps(response[backup[0]]['progress']))
                        if backup[0] in list(self._encryptions.keys()):
                            del self._encryptions[backup[0]]
                    if response[backup[0]]['progress']['status'] == "Success":
                        cluster_level_alerts.on_backup_success(backup[1], backup[2],
                        backup[5])
        return response

    def get_successful_backups_initiated_from_amc(self):

        backups = self.get_successful_backups()
        response = dict(successful_backups=[])
        for backup in backups:
            response['successful_backups'].append(dict(destination_node_address=backup[0],
            destination_location=backup[1], namespace=backup[2], sets=backup[3],
            only_metadata=backup[4]))
        return response

    def get_available_backups(self, source_node_address, source_location,
    username, password):

        source_location = source_location.rstrip("/")
        cmd = "/bin/bash -c 'ls %s | grep backup_'" % source_location
        return management.handle_remote_command(source_node_address, username,
            password, cmd, background=0, backup_flag=False, module_name="shell",
            available_backups=True)

    @synchronous('restore_init_lock')
    def initiate_restore(self, source_node_address, source_location, username,
    password, aerospike_build, threads="20", namespace="",
    missing_records_only="False", ignore_generation_num="False", usr=None, passwd=None, pl_password=None):

        if self._restore == {}:
            optional_args = ""
            if threads != "20":
                optional_args += "-t %s " % threads
            if missing_records_only == "True":
                optional_args += "-u "
            if ignore_generation_num == "True":
                optional_args += "-g "
            if namespace != "":
                optional_args += "-n %s " % namespace
            if usr and passwd:
                optional_args += "-U %s -P%s" % (usr, pl_password)
            restore_id = str(uuid.uuid4())
            progress_file = "/tmp/restore-" + restore_id + ".log"
            source_location = source_location.rstrip("/")

            if aerospike_build.startswith('3'):
                cmd = "/bin/bash -c 'asrestore -h %s -p %s -d %s %s &> %s;'" % (self._host, self._port, source_location, optional_args, progress_file)
            else:
                cmd = "/bin/bash -c 'clrestore2 -h %s -p %s -d %s %s &> %s;'" % (self._host, self._port, source_location, optional_args, progress_file)

            result = management.handle_remote_command(source_node_address,
            username, password, cmd, background=3600, backup_flag=True,
            module_name="shell")

            if result['status'] != "Failure":
                self._restore['source_node_address'] = source_node_address
                self._restore['source_location'] = source_location
                self._restore['namespace'] = namespace
                self._restore['progress_file'] = progress_file
                self._restore['timestamp'] = None
                self._restore['result'] = None
                if username != "" and password is not None:
                    self._restore['encryption'] = TwoWayEncryption()
                    username = self._restore['encryption'].encrypt(str(username))
                    password = self._restore['encryption'].encrypt(str(password))
                self._restore['username'] = username
                self._restore['password'] = password
            return result
        return dict(status="Failure", error="A restore operation on this cluster is already in progress")

    @CachedMethod()
    @synchronous('restore_progress_lock')
    def get_restore_progress(self, aerospike_build, cluster_level_alerts):

        if self._restore != {}:
            if self._restore['result'] != None:
                timestamp = datetime.datetime.now() - datetime.timedelta(seconds=BACKUP_AND_RESTORE_RESULT_HISTORY_TIME)
                timestamp = timestamp.strftime("%Y/%m/%d %H:%M:%S")
                result = self._restore['result']
                if timestamp >= self._restore['timestamp']:
                    self._restore = {}
                return result
            username = self._restore['username']
            password = self._restore['password']
            if username != "" and password is not None:
                username = self._restore['encryption'].decrypt(str(username))
                password = self._restore['encryption'].decrypt(str(password))
            result = management.get_restore_progress(
            self._restore['source_node_address'], username, password,
            self._restore['progress_file'], aerospike_build)
            if result['status'] == "Success":
                cluster_level_alerts.on_restore_success(self._restore['namespace'],
                    self._restore['source_node_address'],
                    self._restore['source_location']
                    )
            if result['status'] == "Success" or result['status'] == "Failure":
                self._restore['result'] = result
                self._restore['timestamp'] = datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S")
            return result
        return dict(status="")
